use entity::master_pass;
use once_cell::sync::Lazy;
use rand::{rngs::OsRng, RngCore};
use scrypt::{scrypt, Params};
use subtle::ConstantTimeEq;

pub const HASH_LENGTH: usize = 64;
pub const SALT_LENGTH: usize = 64;

static PARAMS: Lazy<Params> = Lazy::new(|| Params::new(14, 8, 1, HASH_LENGTH).unwrap());

/// Hashes the bytes with Scrypt with the given salt
#[inline]
pub fn hash_scrypt(bytes: &[u8], salt: &[u8]) -> [u8; HASH_LENGTH] {
    let mut hash = [0; HASH_LENGTH];
    scrypt(bytes, salt, &PARAMS, &mut hash).unwrap();
    hash
}

/// Verifieble scrypt hashed bytes
pub struct HashedBytes<T, U>
where
    T: AsRef<[u8]>,
    U: AsRef<[u8]>,
{
    pub hash: T,
    pub salt: U,
}

impl HashedBytes<[u8; HASH_LENGTH], [u8; SALT_LENGTH]> {
    #[inline]
    pub fn new(bytes: &[u8]) -> Self {
        let mut salt = [0; 64];
        OsRng.fill_bytes(&mut salt);
        Self {
            hash: hash_scrypt(bytes, &salt),
            salt,
        }
    }
}

impl<T, U> HashedBytes<T, U>
where
    T: AsRef<[u8]>,
    U: AsRef<[u8]>,
{
    #[inline]
    pub fn verify(&self, bytes: &[u8]) -> bool {
        let hash = hash_scrypt(bytes, self.salt.as_ref());
        hash.ct_eq(self.hash.as_ref()).into()
    }
}

impl<'a> From<&'a master_pass::Model> for HashedBytes<&'a [u8], &'a [u8]> {
    #[inline]
    fn from(value: &'a master_pass::Model) -> Self {
        HashedBytes {
            hash: &value.password_hash,
            salt: &value.salt,
        }
    }
}

impl From<master_pass::Model> for HashedBytes<Vec<u8>, Vec<u8>> {
    fn from(value: master_pass::Model) -> Self {
        Self {
            hash: value.password_hash,
            salt: value.salt,
        }
    }
}
