use crate::prelude::*;
use futures::future::BoxFuture;

/// A generic state handler. It checks for "/cancel" messages and runs the provided validation function
#[inline]
pub async fn generic<F>(
    bot: Throttle<Bot>,
    msg: Message,
    db: DatabaseConnection,
    dialogue: MainDialogue,
    check: F,
    no_text_message: impl Into<String>,
    next: PackagedHandler<String>,
) -> crate::Result<()>
where
    for<'a> F: FnOnce(
        &'a Message,
        &'a DatabaseConnection,
        &'a str,
    ) -> BoxFuture<'a, crate::Result<Option<String>>>,
{
    let mut handler = next.lock().await;
    if handler.func.is_none() {
        let _ = dialogue.exit().await;
        return Err(HandlerUsed.into());
    }

    let text = match msg.text() {
        Some(text) => text.trim(),
        None => {
            handler
                .previous
                .alter_message(&bot, no_text_message, None, None)
                .await?;
            return Ok(());
        }
    };

    if text == "/cancel" {
        dialogue.exit().await?;
        handler
            .previous
            .alter_message(&bot, "Successfully cancelled", deletion_markup(), None)
            .await?;
        return Ok(());
    }

    if let Some(text) = check(&msg, &db, text).await? {
        handler
            .previous
            .alter_message(&bot, text, None, None)
            .await?;
        return Ok(());
    }

    let previous = handler.previous;
    let func = handler.func.take().unwrap();
    drop(handler);
    let text = text.to_owned();

    if let Err(err) = func(bot, msg, db, dialogue.clone(), previous, text).await {
        let _ = dialogue.exit().await;
        return Err(err);
    }
    Ok(())
}
